use std::fs;

use std::error::Error;
use std::io::{self, ErrorKind, Seek as ioSeek, SeekFrom};

use crate::lws_vfs::{FileInfo, Fstat};

use std::sync::{Arc, Mutex};
use std::collections::HashMap;

extern crate winapi;
use std::os::windows::fs::MetadataExt;
use std::os::windows::prelude::*;
use winapi::um::winnt::FILE_ATTRIBUTE_READONLY;

fn from_metadata(sta: &mut Fstat, metadata: &fs::Metadata) -> Result<(), Box<dyn Error>> {
    sta.fst_size = metadata.len();
    // sta.fst_mtime = metadata
    //             .modified()?.duration_since(SystemTime::UNIX_EPOCH)
    //             .map(|d| d.as_secs() as u64)?;
    sta.fst_mtime = metadata.last_write_time();
    sta.fst_ctime = metadata.creation_time();
    sta.fst_atime = metadata.last_access_time();
    Ok(())
}


#[derive(Debug, Default)]
struct FileFdMgnt {
    // key is fd, value is FileHandle
    files: HashMap<u64, File>,
    curr_fd: u64,
}

impl FileFdMgnt {
    pub fn new() -> FileFdMgnt {
        FileFdMgnt {
            curr_fd: 3,
            files: HashMap::new(),
        }
    }
    fn gen_fd(&mut self) -> Result<u64, Box<dyn Error>> {
        let refd = self.curr_fd;
        loop {
            self.curr_fd += 1;
            match self.files.get(&self.curr_fd) {
                Some(_) => {}
                None => {
                    break;
                }
            }
            if refd == self.curr_fd {
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "no fd",
                )));
            }
        }
        Ok(self.curr_fd)
    }
    pub fn push(&mut self, handle: File) -> Result<u64, Box<dyn Error>> {
        let fd = self.gen_fd()?;
        self.files.insert(fd, handle);
        Ok(fd)
    }
    pub fn pop(&mut self, fd: u64) -> Option<File> {
        self.files.remove(&fd)
    }
}
pub struct FSImpl{
    file_fds: Arc<Mutex<FileFdMgnt>>,
}

impl FSImpl {
pub fn fgetattr(&self, path: &String, sta: &mut Fstat, _fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
    let metadata = fs::metadata(path)?;
    from_metadata(sta, &metadata)?;
    let perm = get_file_mode_mask(path);
    sta.fst_mode = perm as u64;
    Ok(0)
}

pub fn fopen(&self, path: &String, _fi: &mut FileInfo) -> Result<fs::File, Box<dyn Error>> {
    return Ok(fs::File::open(path)?);
}

pub fn fread(
    &self, path: &String,
    buffer: &mut Vec<u8>,
    _size: &mut usize,
    offsize: usize,
    _fi: &mut FileInfo,
) -> Result<i32, Box<dyn Error>> {
    let mut file = fs::File::open(path)?;
    match file.seek_read(buffer, offsize as u64) {
        Ok(size) => {
            println!(
                "size is:{}, buffer is {}",
                size,
                String::from_utf8(buffer.to_vec()).unwrap()
            );
            buffer.resize(size, 0)
        }
        Err(_) => {
            buffer.resize(file.stream_position()? as usize - offsize, 0);
        }
    }
    Ok(0)
}

pub fn fwrite(
    &self, path: &String,
    buffer: &Vec<u8>,
    size: &mut usize,
    offsize: usize,
    _fi: &mut FileInfo,
) -> Result<i32, Box<dyn Error>> {
    let mut file = fs::File::options().write(true).create(true).open(path)?;
    match file.seek_write(buffer, offsize as u64) {
        Ok(written) => {
            *size = written as usize;
        }
        Err(_) => {
            *size = file.stream_position()? as usize - offsize;
        }
    }
    Ok(0)
}
pub fn freaddir(
    &self, path: &String,
    buffer: &mut Vec<String>,
    _size: usize,
    _offset: usize,
    _fi: &mut FileInfo,
) -> Result<i32, Box<dyn Error>> {
    let mut readed = fs::read_dir(path)?;
    loop {
        match readed.next() {
            Some(Ok(entry)) => {
                let name = entry.file_name().into_string().unwrap();
                buffer.push(name);
            }
            Some(Err(e)) => {
                return Err(Box::new(e));
            }
            None => {
                break;
            }
        }
    }
    Ok(0)
}
pub fn fmkdir(&self, path: &String, _mode: u32) -> Result<i32, Box<dyn Error>> {
    fs::create_dir(path)?;
    Ok(0)
}
pub fn fchown(&self, _path: &String, _uid: u32, _gid: u32) -> Result<i32, Box<dyn Error>> {
    Ok(-1)
}
pub fn ftruncate(&self, path: &String, size: u64) -> Result<i32, Box<dyn Error>> {
    let mut file = fs::File::options().write(true)
        .truncate(true)
        .open(path)?;
    file.seek(SeekFrom::Start(size))?;
    file.set_len(size)?;
    Ok(0)
}

pub fn futimens(&self, path: &String, a: &Vec<u64>, m: &Vec<u64>) -> Result<i32, Box<dyn Error>> {
    use std::ptr::null_mut;
    use winapi::ctypes::c_void;
    use winapi::shared::minwindef::DWORD;
    use winapi::shared::minwindef::FILETIME;
    use winapi::um::fileapi::SetFileTime;
    fn system_time_to_file_time(seconds: u64, nanoseconds: u32) -> FILETIME {
        let total_nanoseconds = (seconds * 1_000_000_000) + nanoseconds as u64;
        let total_100_nanoseconds = total_nanoseconds / 100;
        FILETIME {
            dwLowDateTime: total_100_nanoseconds as DWORD,
            dwHighDateTime: (total_100_nanoseconds >> 32) as DWORD,
        }
    }
    let file = fs::File::options().write(true).open(path)?;
    let handle = file.as_raw_handle();

    let atime = system_time_to_file_time(a[0], a[1] as u32);
    let mtime = system_time_to_file_time(m[0], m[1] as u32);

    let result = unsafe { SetFileTime(handle as *mut c_void, null_mut(), &atime, &mtime) };

    if result != 0 {
        Ok(0)
    } else {
        Err(Box::new(io::Error::last_os_error()))
    }
}
pub fn frelease(&self, _path: &String, _fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
    if let Some(fh) == self.file_fds.lock().unwrap().pop(fi.fh){
        drop(fh);
        Ok(0)
    } else {
        Ok(-1)
    }
}
pub fn fsetxattr(
    &self, _path: &String,
    _name: &str,
    _value: &Vec<u8>,
    _size: usize,
    _flags: u32,
) -> Result<i32, Box<dyn Error>> {
    Ok(-1)
}
pub fn fgetxattr(&self, _path: &String, _name: &str, _size: usize) -> Result<i32, Box<dyn Error>> {
    Ok(-1)
}

fn get_file_mode_mask(path: &str) -> u32 {
    let mut permissions = 0;
    let metadata = match fs::metadata(path) {
        Ok(metadata) => metadata,
        Err(_) => {
            println!("get {} metadata fail", path);
            return permissions;
        }
    };
    let attributes = metadata.file_attributes();
    // let permissions = metadata.permissions();
    if attributes & FILE_ATTRIBUTE_READONLY != 0 {
        permissions |= 0o222;
    } else {
        permissions |= 0o444 | 0o222;
    }
    if metadata.is_dir() {
        permissions |= 0o111;
    }
    permissions
}
pub fn faccess(&self, path: &String, mask: u32) -> Result<i32, Box<dyn Error>> {
    let permissions = get_file_mode_mask(path);
    Ok(if permissions & mask != 0 { 0 } else { -1 })
}

pub fn frmdir(&self, path: &String) -> Result<i32, Box<dyn Error>> {
    fs::remove_dir(path)?;
    Ok(0)
}
}

#[test]
fn test_fread() {
    use crate::fs_impl::*;
    let path = String::from("test_fread.txt");
    let file: fs::File = fs::File::options()
        .write(true)
        .create(true)
        .open(&path)
        .unwrap();

    let ref_buffer = b"This is a test file.";
    let _ = file.seek_write(ref_buffer, 0);

    let mut fi = FileInfo::default();

    //read from file middle
    let mut size = 4;
    let mut offsize: usize = 5;
    let mut buffer = vec![0; size];
    let _ = fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 4);
    assert_eq!(buffer, b"is a");

    //read from file head
    offsize = 0;
    let _ = fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 4);
    assert_eq!(buffer, b"This");

    // read from the file tail, read len > file remain
    offsize = ref_buffer.len() - 2;
    let _ = fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 2);
    assert_eq!(buffer, b"e.");

    // cleanup
    let _ = fs::remove_file(path);
}
#[test]
fn test_fwrite() {
    use crate::fs_impl::*;
    let path = String::from("test_fwrite.txt");
    let _ = fs::remove_file(&path);
    let mut buffer = Vec::from("This is a test file");
    let mut size = buffer.len();

    //write from file head
    let mut offsize: usize = 0;
    let mut fi = FileInfo::default();
    let _ = fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());

    //write from file middle
    offsize = 5;
    let _ = fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());
    let file_metadata = fs::metadata(&path).unwrap();
    let file_size = file_metadata.len();
    assert_eq!(file_size, (buffer.len() + offsize) as u64);

    //write from file ended
    offsize = file_size as usize + 10;
    let _ = fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());
    let file_metadata = fs::metadata(&path).unwrap();
    let file_size = file_metadata.len();
    assert_eq!(file_size, (buffer.len() + offsize) as u64);

    // cleanup
    //let _ = fs::remove_file(path);
}
#[test]
fn test_dir() {
    use std::path::Path;
    let path = String::from("test_dir");
    let sub_path = String::from("test_dir/test_sub_dir");
    //创建父目录
    let _ = fmkdir(&path, 0);
    //创建子目录
    let _ = fmkdir(&sub_path, 0);
    assert_eq!(true, Path::new(&sub_path).is_dir());
    assert_eq!(true, Path::new(&path).is_dir());
    //一次性创建多级目录
    let ret = fmkdir(&String::from("patent/child"), 0);
    assert_ne!(
        match ret {
            Ok(ret) => ret,
            Err(_) => -1,
        },
        0
    );
    fs::remove_dir_all(path);
}
// let result = fwrite(&path, &buffer, &mut size, offsize, &mut fi);
// println!("result: {:?}", result);
// println!("size: {:?}", size);
// println!("buffer: {:?}", buffer);
// let result = freaddir(&path, &mut vec![], 0, 0, &mut fi);
// println!("result: {:?}", result);
// println!("buffer: {:?}", buffer);
// let result = fmkdir(&path, 0);
// println!("result: {:?}", result);
// let result = fchown(&path, 0, 0);
// println!("result: {:?}", result);
// let result = ftruncate(&path, 100);
// println!("result: {:?}", result);
// let result = futimens(&path, &vec![100, 0], &vec![100, 0]);
// println!("result: {:?}", result);
// let result = frelease(&path, &mut fi);
// println!("result: {:?}", result);
// let result = fsetxattr(&path, "test", &vec![1,2,3], 3, 0);
// println!("result: {:?}", result);
// let result = fgetxattr(&path, "test", 100);
// println!("result: {:?}", result);
// let result = faccess(&path, 0o777);
// println!("result: {:?}", result);
