use lws_vfs::lws_vfs_server::LwsVfs;
use lws_vfs::{
    Access, Chown, FileInfo, Flush, Fstat, Getattr, Getxattr, HelloReply, HelloRequest, Mkdir,
    Open, Read, Readdir, Release, Rmdir, Setxattr, Truncate, Utimens, Write, Flush,
};
use serde_json::{self, Value};
use std::error::Error;
use std::fs::File;
use std::io::{BufReader, Read as _};
use tonic::{Request, Response, Status};
mod fs_impl;
pub mod lws_vfs {
    tonic::include_proto!("lws_vfs");
}

// config
//
// {
//     "mount": {
//         "c:\\": "/l3c",
//         "w:\\": "/l0e",
//         "y:\\": "/l0d"
//     },
//     "port":5001
// }
//
#[derive(Debug, Default)]
pub struct Config {
    port: u16,
    mount_map: HashMap<String, String>,
}
impl Config {
    pub fn new(json: &str) -> Result<Config, Box<dyn Error>> {
        let mut file = File::open(json)?;
        let mut buffer = String::new();
        file.read_to_string(&mut buffer)?;
        let json: Value = serde_json::from_str(buffer.as_ref())?;
        let port: u16 = match json.get("port") {
            Some(port) => port.as_u64().expect("expect port is a number but its not") as u16,
            None => 5001,
        };
        let mounts = match json.get("mount") {
            Some(mounts) => mounts.as_object().unwrap(),
            None => {
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "no mount map",
                )))
            }
        };
        let mount_map = mounts
            .iter()
            .map(|(key, value)| {
                (
                    key.as_str().to_string(),
                    value.as_str().unwrap().to_string(),
                )
            })
            .collect();
        Ok(Config { port, mount_map })
    }

    pub fn get_port(&self) -> u16 {
        self.port
    }
}

/// server 的基础结构
#[derive(Debug, Default)]
pub struct LwsVfsIns {
    pub config: Config,
    file_fds: Arc<Mutex<FileFdMgnt>>,
}

impl LwsVfsIns {
    pub fn new(json: &str) -> Result<LwsVfsIns, Box<dyn Error>> {
        Ok(LwsVfsIns {
            config: Config::new(json)?,
            file_fds: Arc::new(Mutex::new(FileFdMgnt::new())),
        })
    }
    fn lpath(&self, path: &String) -> String {
        let mut ret = String::new();
        for (k, v) in &self.config.mount_map {
            if path.starts_with(v) {
                ret = path.replace(v, k);
                break;
            }
        }
        ret
    }
}

#[tonic::async_trait]
impl LwsVfs for LwsVfsIns {
    async fn say_hello(
        &self,
        request: Request<HelloRequest>,
    ) -> Result<Response<HelloReply>, Status> {
        println!("Got a request: {:?}", request);

        let reply = HelloReply {
            message: format!("Hello {}!", request.into_inner().name),
        };

        Ok(Response::new(reply))
    }
    async fn fgetattr(&self, request: Request<Getattr>) -> Result<Response<Getattr>, Status> {
        println!("Got a request: {:?}", request);

        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mut fstat = Fstat::default();
        let mut fi = FileInfo::default();
        let ret = match fs_impl::fgetattr(path, &mut fstat, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error getting file metadata: {:?}", e);
                -1
            }
        };
        let reply = Getattr {
            path: path.to_string(),
            stat: Some(fstat),
            ret,
        };
        Ok(Response::new(reply))
    }
    async fn fopen(&self, request: Request<Open>) -> Result<Response<Open>, Status> {
        println!("Got a request: {:?}", request);
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mut fi = request.fi.unwrap();
        let ret = match fs_impl::fopen(path, &mut fi) {
            Ok(file) => {
                match self.file_fds.lock().unwrap().push(file) {
                    Ok(fd) => fi.fh = fd,
                    Err(e) => {
                        println!("Error pushing file[{}]: {:?}", path, e);
                    }
                };
                0
            }
            Err(e) => {
                println!("Error opening file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Open {
            path: path.to_string(),
            fi: Some(fi),
            ret,
        };
        Ok(Response::new(reply))
    }
    async fn fread(&self, request: Request<Read>) -> Result<Response<Read>, Status> {
        println!("Got a request: {:?}", request);
        let request = request.into_inner();
        let mut fi = request.fi.unwrap();
        let path = &self.lpath(&request.path);
        let mut size: usize = request.size.try_into().unwrap();
        let offset: usize = request.offset.try_into().unwrap();
        let mut buff: Vec<u8> = Vec::with_capacity(size);
        unsafe {
            buff.set_len(size);
        }
        match fs_impl::fread(path, &mut buff, &mut size, offset, &mut fi) {
            Ok(ret) => {
                let reply = Read {
                    path: request.path,
                    size: buff.len() as i64,
                    buff,
                    offset: offset as i64,
                    fi: Some(fi),
                    ret,
                };
                Ok(Response::new(reply))
            }
            Err(e) => {
                println!("Error reading file[{}]: {:?}", path, e);
                let reply = Read {
                    path: request.path,
                    buff: Vec::new(),
                    size: 0,
                    offset: 0,
                    fi: Some(fi),
                    ret: -1,
                };
                Ok(Response::new(reply))
            }
        }
    }
    async fn fwrite(&self, request: Request<Write>) -> Result<Response<Write>, Status> {
        let request = request.into_inner();
        let mut fi = request.fi.unwrap();
        let path = &self.lpath(&request.path);
        let buff = request.buff;
        let mut size = request.size as usize;
        let offset = request.offset as usize;
        let ret = match fs_impl::fwrite(path, &buff, &mut size, offset, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error fwrite file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Write {
            ret,
            size: size as i64,
            ..Default::default()
        };
        Ok(Response::new(reply))
    }
    async fn faccess(&self, request: Request<Access>) -> Result<Response<Access>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mask = request.mask;
        let ret = match fs_impl::faccess(path, mask) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error access file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Access {
            ret,
            path: path.to_string(),
            mask,
        };
        Ok(Response::new(reply))
    }

    async fn fsetxattr(&self, request: Request<Setxattr>) -> Result<Response<Setxattr>, Status> {
        let req = request.into_inner();
        let path = &self.lpath(&req.path);
        let name = req.name.as_ref();
        let value = req.value;
        let size = req.size as usize;
        let flags = req.flags;
        let ret = match fs_impl::fsetxattr(path, name, &value, size, flags) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: setxattr file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Setxattr {
            ret: ret,
            ..Setxattr::default()
        };
        Ok(Response::new(reply))
    }
    async fn fgetxattr(&self, request: Request<Getxattr>) -> Result<Response<Getxattr>, Status> {
        let req = request.into_inner();
        let path = &self.lpath(&req.path);
        let name = req.name.as_ref();
        let size = req.size as usize;
        let ret = match fs_impl::fgetxattr(path, name, size) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: getxattr file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Getxattr {
            ret,
            ..Getxattr::default()
        };
        Ok(Response::new(reply))
    }

    async fn freaddir(&self, request: Request<Readdir>) -> Result<Response<Readdir>, Status> {
        let request = request.into_inner();
        let mut fi = request.fi.unwrap();
        let path = &self.lpath(&request.path);
        let offset = request.offset as usize;
        let mut dirs: Vec<String> = Vec::new();
        let size = 0;
        let ret = match fs_impl::freaddir(path, &mut dirs, size, offset, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: readdir file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Readdir {
            ret,
            dirs,
            ..Readdir::default()
        };
        Ok(Response::new(reply))
    }
    async fn fmkdir(&self, request: Request<Mkdir>) -> Result<Response<Mkdir>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mode = request.mode;
        let ret = match fs_impl::fmkdir(&path, mode) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: mkdir [{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Mkdir {
            ret,
            ..Mkdir::default()
        };
        Ok(Response::new(reply))
    }
    async fn ftruncate(&self, request: Request<Truncate>) -> Result<Response<Truncate>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let size = request.size as u64;
        let ret = match fs_impl::ftruncate(path, size) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: truncate file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Truncate {
            ret,
            ..Truncate::default()
        };
        Ok(Response::new(reply))
    }

    async fn futimens(&self, request: Request<Utimens>) -> Result<Response<Utimens>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let a_time = vec![request.ts[0].tv_sec as u64, request.ts[0].tv_nsec as u64];
        let m_time = vec![request.ts[1].tv_sec as u64, request.ts[1].tv_nsec as u64];
        let ret = match fs_impl::futimens(path, &a_time, &m_time) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: futimens file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Utimens {
            ret,
            ..Utimens::default()
        };
        Ok(Response::new(reply))
    }

    async fn fchown(&self, request: Request<Chown>) -> Result<Response<Chown>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let uid = request.uid as u32;
        let gid = request.gid as u32;
        let ret = match fs_impl::fchown(path, uid, gid) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: chown file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Chown {
            ret,
            ..Chown::default()
        };
        Ok(Response::new(reply))
    }
    async fn frelease(&self, request: Request<Release>) -> Result<Response<Release>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mut fi = request.fi.unwrap();
        self.file_fds.lock().unwrap().pop(fi.fh);
        let ret = match fs_impl::frelease(path, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: release file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Release {
            ret,
            ..Release::default()
        };
        Ok(Response::new(reply))
    }
    async fn frmdir(&self, request: Request<Rmdir>) -> Result<Response<Rmdir>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let ret = match fs_impl::frmdir(path) {
            Ok(ret) => ret,
            Err(e) => {
                println!("Error: rmdir [{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Rmdir {
            ret,
            path: request.path,
        };
        Ok(Response::new(reply))
    }

    async fn fflush(&self, request: Request<Flush>) -> Result<Response<Flush>, Status> {
        let request = request.into_inner();
        let path = &self.lpath(&request.path);
        let mut fi = request.fi.unwrap();
        self.file_fds.lock().unwrap().pop(fi.fh);
    }
}
